#!/usr/bin/env python3

from dgenies import MODE

import os
import shutil
from math import sqrt
from numpy import mean
from pathlib import Path
import json
from dgenies.bin.index import Index
from dgenies.config_reader import AppConfigReader
from dgenies.lib.functions import Functions
from intervaltree import IntervalTree
import matplotlib as mpl
mpl.use('Agg')
from matplotlib import pyplot as plt
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord


class Paf:
    """
    Functions applied to PAF files
    """
    config = AppConfigReader()
    limit_idy = [0.25, 0.5, 0.75]
    max_nb_lines = config.max_nb_lines

    def __init__(self, paf: str, idx_q: str, idx_t: str, auto_parse: bool=True, mailer=None, id_job=None):
        """

        :param paf: PAF file path
        :type paf: str
        :param idx_q: query index file path
        :type idx_q: str
        :param idx_t: target index file path
        :type idx_t: str
        :param auto_parse: if True, parse PAF file at initialisation
        :type auto_parse: bool
        :param mailer: mailer object, to send mails
        :type mailer: Mailer
        :param id_job: job id
        :type id_job: str
        """
        self.paf = paf
        self.idx_q = idx_q
        self.idx_t = idx_t
        self.sorted = False
        self.data_dir = os.path.dirname(paf)
        if os.path.exists(os.path.join(self.data_dir, ".sorted")):
            self.paf += ".sorted"
            self.idx_q += ".sorted"
            self.sorted = True
        self.sampled= False
        self.len_q = None
        self.len_t = None
        self.min_idy = None
        self.max_idy = None
        self.lines = {}
        self.q_contigs = {}
        self.q_order = []
        self.t_contigs = {}
        self.t_order = []
        self.q_reversed = {}
        self.name_q = None
        self.name_t = None
        self.parsed = False
        self.error = False
        self.mailer = mailer
        self.id_job = id_job
        self.q_abs_start = {}
        self.t_abs_start = {}

        if auto_parse:
            self.parse_paf()

    @staticmethod
    def _flush_blocks(index_c, new_index_c, new_index_o, current_block):
        """
        When parsing index, build a mix of too small sequential contigs (if their number exceed 5), else just add
        co to the new index

        :param index_c: current index contigs def
        :type index_c: dict
        :param new_index_o: new index contigs order
        :type new_index_o: list
        :param new_index_c: new index contigs def
        :type new_index_c: dict
        :param current_block: contigs in the current analyzed block
        :type current_block: list
        :return: (new index contigs defs, new index contigs order)
        :rtype: (dict, list)
        """
        if len(current_block) >= 5:
            block_length = 0
            for contig in current_block:
                block_length += index_c[contig]
            b_name = "###MIX###_" + "###".join(current_block)
            new_index_c[b_name] = block_length
            new_index_o.append(b_name)
        elif len(current_block) > 0:
            for b_name in current_block:
                new_index_c[b_name] = index_c[b_name]
                new_index_o.append(b_name)
        return new_index_c, new_index_o

    def parse_index(self, index_o: list, index_c: dict, full_len: int):
        """
        Parse index and merge too small contigs together

        :param index_o: index contigs order
        :type index_o: list
        :param index_c: index contigs def
        :type index_c: dict
        :param full_len: length of the sequence
        :type full_len: int
        :return: (new contigs def, new contigs order)
        :rtype: (dict, list)
        """
        new_index_o = []
        new_index_c = {}
        current_block = []
        for index in index_o:
            if index_c[index] >= 0.002 * full_len:
                new_index_c, new_index_o = self._flush_blocks(index_c, new_index_c, new_index_o, current_block)
                current_block = []
                new_index_c[index] = index_c[index]
                new_index_o.append(index)
            else:
                current_block.append(index)
        new_index_c, new_index_o = self._flush_blocks(index_c, new_index_c, new_index_o, current_block)
        return new_index_c, new_index_o

    @staticmethod
    def remove_noise(lines, noise_limit):
        """
        Remove noise from the dot plot

        :param lines: lines of the dot plot, by class
        :type lines: dict
        :param noise_limit: line length limit
        :type noise_limit: float
        :return: kept lines, by class
        :rtype: dict
        """
        keep_lines = {
            "0": [],
            "1": [],
            "2": [],
            "3": []
        }
        for cls, c_lines in lines.items():
            for line in c_lines:
                x1 = line[0]
                x2 = line[1]
                y1 = line[2]
                y2 = line[3]
                idy = line[4]
                len_m = sqrt(pow(x2 - x1, 2) + pow(y2 - y1, 2))
                if len_m >= noise_limit:
                    keep_lines[cls].append(line)
        return keep_lines

    def keyerror_message(self, exception, type_f):
        """
        Build message if contig not found in query or target

        :param exception: exception object
        :type exception: KeyError
        :param type_f: type of data (query or target)
        :type type_f: str
        :return: error message
        :rtype: str
        """
        message = "Invalid contig for %s: %s" % (type_f, exception.args[0])
        if os.path.exists(os.path.join(self.data_dir, ".align")):
            message += ". May be you invert query and target files?"
        return message

    def parse_paf(self, merge_index=True, noise=True):
        """
        Parse PAF file

        :param merge_index: if True, merge too small contigs in index
        :type merge_index: bool
        :param noise: if True, remove noise
        :type noise: bool
        """
        min_idy = 10000000000
        max_idy = -10000000000
        lines = {
            "0": [],  # idy < 0.25
            "1": [],  # idy < 0.5
            "2": [],  # idy < 0.75
            "3": []  # idy > 0.75
        }
        try:
            name_q, q_order, q_contigs, q_reversed, q_abs_start, len_q = Index.load(self.idx_q)
            self.q_abs_start = q_abs_start
            if merge_index:
                q_contigs, q_order = self.parse_index(q_order, q_contigs, len_q)
        except IOError:
            self.error = "Index file does not exist for query!"
            return False

        try:
            name_t, t_order, t_contigs, t_reversed, t_abs_start, len_t = Index.load(self.idx_t)
            self.t_abs_start = t_abs_start
            if merge_index:
                t_contigs, t_order = self.parse_index(t_order, t_contigs, len_t)
        except IOError:
            self.error = "Index file does not exist for target!"
            return False

        lines_lens = []

        try:
            with open(self.paf, "r") as paf_file:
                nb_lines = 0
                for line in paf_file:
                    nb_lines += 1
                    if nb_lines > self.max_nb_lines:
                        self.sampled = True
                        break
                    parts = line.strip("\n").split("\t")
                    v1 = parts[0]
                    v6 = parts[5]
                    strand = 1 if parts[4] == "+" else -1
                    idy = int(parts[9]) / int(parts[10])
                    min_idy = min(min_idy, idy)
                    max_idy = max(max_idy, idy)
                    # x1, x2, y1, y2, idy
                    try:
                        y1 = int(parts[2]) + q_abs_start[v1]
                        y2 = int(parts[3]) + q_abs_start[v1]
                    except KeyError as e:
                        self.error = self.keyerror_message(e, "query")
                        return False
                    try:
                        x1 = int(parts[7 if strand == 1 else 8]) + t_abs_start[v6]
                        x2 = int(parts[8 if strand == 1 else 7]) + t_abs_start[v6]
                    except KeyError as e:
                        self.error = self.keyerror_message(e, "target")
                        return False
                    len_m = sqrt(pow(x2 - x1, 2) + pow(y2 - y1, 2))
                    lines_lens.append(len_m)
                    if idy < self.limit_idy[0]:
                        class_idy = "0"
                    elif idy < self.limit_idy[1]:
                        class_idy = "1"
                    elif idy < self.limit_idy[2]:
                        class_idy = "2"
                    else:
                        class_idy = "3"
                    lines[class_idy].append([x1, x2, y1, y2, idy, v1, v6])
        except IOError:
            self.error = "PAF file does not exist!"
            return False

        if not noise and nb_lines > 1000:
            counts, bins, bars = plt.hist(lines_lens, bins=nb_lines//10)
            counts = list(counts)
            max_value = max(counts)
            max_index = counts.index(max_value)
            limit_index = -1
            for i in range(max_index, len(counts)):
                if counts[i] < max_value / 50:
                    limit_index = i
                    break
            if limit_index > -1:
                lines = self.remove_noise(lines, bins[limit_index])

        self.parsed = True
        self.len_q = len_q
        self.len_t = len_t
        self.min_idy = min_idy
        self.max_idy = max_idy
        self.lines = lines
        self.q_contigs = q_contigs
        self.q_order = q_order
        self.q_reversed = q_reversed
        self.t_contigs = t_contigs
        self.t_order = t_order
        self.name_q = name_q
        self.name_t = name_t

    def get_d3js_data(self):
        """
        Build data for D3.js client

        :return: data for d3.js:

            * y_len: length of query (Bp)
            * x_len: length of target (Bp)
            * min_idy: minimum of identity (float)
            * max_idy: maximum of identity (float)
            * lines: matches lines, by class of identity (dict)
            * y_contigs: query contigs definitions (dict)
            * y_order: query contigs order (list)
            * x_contigs: target contigs definitions (dict)
            * x_order: target contigs order (list)
            * name_y: name of the query (str)
            * name_x: name of the target (str)
            * limit_idy: limit for each class of identities (list)
        :rtype: dict
        """
        return {
            'y_len': self.len_q,
            'x_len': self.len_t,
            'min_idy': self.min_idy,
            'max_idy': self.max_idy,
            'lines': self.lines,
            'y_contigs': self.q_contigs,
            'y_order': self.q_order,
            'x_contigs': self.t_contigs,
            'x_order': self.t_order,
            'name_y': self.name_q,
            'name_x': self.name_t,
            'limit_idy': self.limit_idy,
            'sorted': self.sorted,
            'sampled': self.sampled,
            "max_nb_lines": self.max_nb_lines,
        }

    def save_json(self, out):
        """
        Save D3.js data to json

        :param out: output file path
        :type out: str
        """
        import json
        data = self.get_d3js_data()
        with open(out, "w") as out_f:
            out_f.write(json.dumps(data))

    def is_contig_well_oriented(self, lines, contig, chrom):
        """
        Returns True if the contig is well oriented. A well oriented contig
        must have y increased when x increased. We check that only for highest matches
        (small matches must be ignored)

        :param lines: lines inside the contig
        :type lines: list
        :param contig: query contig name
        :type contig: str
        :param chrom: target chromosome name
        :type chrom: str
        :return: True if well oriented, False else
        :rtype: bool
        """
        lines.sort(key=lambda x: -x[-1])
        max_len = lines[0][-1]
        i = 0

        # Select samples of tested lines such that:
        # - a selected line is at least 10% of the longest line
        # - and a selected line is at least 1% of the length of contig or the length of chrom
        while i < len(lines) and lines[i][-1] > max_len * 0.1 \
                and lines[i][-1] >= 0.01 * min(self.q_contigs[contig], self.t_contigs[chrom]):
            i += 1
        selected_lines = lines[:i]

        # Check orientation of each line:
        if len(selected_lines) > 1:
            selected_lines.sort(key=lambda x: x[0])
            orients = []
            for i in range(1, len(selected_lines)):
                if selected_lines[i][2] > selected_lines[i-1][2]:
                    orients.append(1)
                else:
                    orients.append(-1)
            if mean(orients) > -0.1:  # We have a good orientation (-0.1 to ignore ambiguous cases)
                return True
        elif len(selected_lines) == 1:
            orient_first = "+" if selected_lines[0][3] < selected_lines[0][4] else "-"  # Get X orientation
            if selected_lines[0][-2 if orient_first == "+" else -3] > \
                    selected_lines[0][-3 if orient_first == "+" else -2]:  # Check Y according to X orientation
                return True
        elif len(selected_lines) == 0:  # None line were selected: we ignore this block
            return True

        # In all other cases the orientation is wrong:
        return False

    def reorient_contigs_in_paf(self, contigs):
        """
        Reorient contigs in the PAF file

        :param contigs: contigs to be reoriented
        """
        sorted_file = self.paf + ".sorted"
        contigs = set(contigs)
        with open(self.paf, "r") as source, open(sorted_file, "w") as target:
            for line in source:
                parts = line.strip("\n").split("\t")
                if parts[0] in contigs:
                    len_q = int(parts[1])
                    x1_q = int(parts[2])
                    x2_q = int(parts[3])
                    if parts[4] == "-":
                        parts[4] = "+"
                    else:
                        parts[4] = "-"
                    parts[2] = str(len_q - x2_q)
                    parts[3] = str(len_q - x1_q)
                target.write("\t".join(parts) + "\n")
        if self.paf.endswith(".sorted"):
            os.remove(self.paf)
            shutil.move(sorted_file, self.paf)
        else:
            self.paf = sorted_file
        return True

    def _update_query_index(self, contigs_reoriented):
        """
        Write new query index file (including new reoriented contigs info)

        :param contigs_reoriented: reoriented contigs list
        :type contigs_reoriented: list
        """
        contigs_reoriented = set(contigs_reoriented)
        with open(self.idx_q, "w") as idx:
            idx.write(self.name_q + "\n")
            for contig in self.q_order:
                idx.write("\t".join([contig, str(self.q_contigs[contig]), "1" if contig in contigs_reoriented else "0"])
                          + "\n")

    def set_sorted(self, is_sorted):
        """
        Change sorted status

        :param is_sorted: new sorted status
        :type is_sorted: bool
        """
        self.sorted = is_sorted
        sorted_touch = os.path.join(os.path.dirname(self.paf), ".sorted")
        if is_sorted:
            Path(sorted_touch).touch()
        else:
            if os.path.exists(sorted_touch):
                os.remove(sorted_touch)

    def compute_gravity_contigs(self):
        """
        Compute gravity for each contig on each chromosome (how many big matches they have).
        Will be used to find which chromosome has the highest value for each contig

        :return:

            * [0] gravity for each contig and each chromosome:
                {contig1: {chr1: value, chr2: value, ...}, contig2: ...}
            * [1] For each block save lines inside:
                [median_on_query, squared length, median_on_target, x1, x2, y1, y2, length] (x : on target, y: on query)
        """
        gravity_contig = {}
        lines_on_block = {}
        # Compute size of blocks (in term of how many big match they have), and save median of each match on each one
        # (for next step)
        for line in [j for i in list(self.lines.values()) for j in i]:
            x1 = int(line[0])
            x2 = int(line[1])
            y1 = int(line[2])
            y2 = int(line[3])
            contig = line[5]
            chrm = line[6]
            block = (contig, chrm)
            # X1 and X2 (in good orientation):
            x_1 = min(x1, x2)
            x_2 = max(x2, x1)
            med_q = x_1 + (abs(x_2 - x_1) / 2)
            # Y1 and Y2 (in good orientation):
            y_1 = min(y1, y2)
            y_2 = max(y2, y1)
            med_t = y_1 + (abs(y_2 - y_1) / 2)
            len_m = sqrt(pow(x2 - x1, 2) + pow(y2 - y1, 2))  # Len
            len_m_2 = pow(1 + len_m, 2) # Pow of len
            if block not in lines_on_block:
                lines_on_block[block] = []
            lines_on_block[block].append((med_q, len_m_2, med_t, x1, x2, y1, y2, len_m))

            if contig not in gravity_contig:
                gravity_contig[contig] = {}
            if chrm not in gravity_contig[contig]:
                gravity_contig[contig][chrm] = 0
            gravity_contig[contig][chrm] += len_m_2
        return gravity_contig, lines_on_block

    def sort(self):
        """
        Sort contigs according to reference target and reorient them if needed
        """
        self.parse_paf(False)
        sorted_file = self.paf + ".sorted"
        if not self.sorted:  # Do the sort
            if not self.paf.endswith(".sorted") and not self.idx_q.endswith(".sorted") and \
                    (not os.path.exists(self.paf + ".sorted") or not os.path.exists(self.idx_q + ".sorted")):
                gravity_contig , lines_on_block = self.compute_gravity_contigs()

                # For each contig, find best block, and deduce gravity of contig:
                gravity_on_contig = {}
                reorient_contigs = []
                for contig, chr_blocks in gravity_contig.items():
                    # Find best block:
                    max_number = 0
                    max_chr = None
                    for chrm, size in chr_blocks.items():
                        if size > max_number:
                            max_number = size
                            max_chr = chrm

                    # Compute gravity of contig:
                    nb_items = 0
                    sum_items = 0
                    lines_on_selected_block = lines_on_block[(contig, max_chr)]
                    for med in lines_on_selected_block:
                        sum_items += med[0] * med[1]
                        nb_items += med[1]
                    gravity_on_contig[contig] = sum_items / nb_items

                    # Check if contig must be re-oriented:
                    if len(lines_on_selected_block) > 0:
                        if not self.is_contig_well_oriented(lines_on_selected_block, contig, max_chr):
                            reorient_contigs.append(contig)

                # Sort contigs:
                self.q_order.sort(key=lambda x: gravity_on_contig[x] if x in gravity_on_contig else self.len_q + 1000)

                self.idx_q += ".sorted"

                with open(self.idx_q, "w") as idx_q_f:
                    idx_q_f.write(self.name_q + "\n")
                    for contig in self.q_order:
                        idx_q_f.write("\t".join([contig, str(self.q_contigs[contig])]) + "\n")

                # Re-orient contigs:
                if len(reorient_contigs) > 0:
                    self.reorient_contigs_in_paf(reorient_contigs)
                else:
                    shutil.copyfile(self.paf, sorted_file)

                # Update index:
                self._update_query_index(reorient_contigs)

            else:
                self.idx_q += ".sorted"
            self.set_sorted(True)
            self.paf = sorted_file

        else:  # Undo the sort
            self.paf = self.paf.replace(".sorted", "")
            self.idx_q = self.idx_q.replace(".sorted", "")
            self.set_sorted(False)

        # Re parse PAF file:
        self.parsed = False
        self.parse_paf()

    def reverse_contig(self, contig_name):
        """
        Reverse contig

        :param contig_name: contig name
        :type contig_name: str
        """
        self.parse_paf(False)
        reorient_contigs = [contig_name]
        self.reorient_contigs_in_paf(reorient_contigs)
        if not self.idx_q.endswith(".sorted"):
            self.idx_q += ".sorted"
        self._update_query_index(reorient_contigs)
        self.set_sorted(True)
        self.parsed = False
        self.parse_paf()

    def get_query_on_target_association(self, with_coords=True):
        """
        For each query, get the best matching chromosome

        :return: query on target association
        :rtype: dict
        """
        gravity_contig, lines_on_block = self.compute_gravity_contigs()
        query_on_target = {}
        for contig, chr_blocks in gravity_contig.items():
            # Find best block:
            max_number = 0
            max_chr = None
            min_target = -1
            max_target = -1
            min_query = -1
            max_query = -1
            for chrm, size in chr_blocks.items():
                if size > max_number:
                    max_number = size
                    max_chr = chrm
                    min_target = -1
                    max_target = -1
                    min_query = -1
                    max_query = -1
                    if with_coords:
                        for line in lines_on_block[(contig, chrm)]:
                            x1 = min(line[3], line[4]) - self.t_abs_start[chrm]
                            x2 = max(line[3], line[4]) - self.t_abs_start[chrm]
                            if x1 < min_target or min_target == -1:
                                min_target = x1
                            if x2 > max_target:
                                max_target = x2
                            y1 = min(line[5], line[6]) - self.q_abs_start[contig]
                            y2 = max(line[5], line[6]) - self.q_abs_start[contig]
                            if y1 < min_query or min_query == -1:
                                min_query = y1
                            if y2 > max_query:
                                max_query = y2
            if max_chr is not None:
                query_on_target[contig] = (max_chr, min_query, max_query, min_target, max_target)
            else:
                query_on_target[contig] = None
        return query_on_target

    def get_queries_on_target_association(self):
        """
        For each target, get the list of queries associated to it

        :return: list of queries associated to each target
        :rtype: dict
        """
        gravity_contig = self.compute_gravity_contigs()[0]
        queries_on_target = {}
        for contig, chr_blocks in gravity_contig.items():
            # Find best block:
            max_number = 0
            max_chr = None
            for chrm, size in chr_blocks.items():
                if size > max_number:
                    max_number = size
                    max_chr = chrm
            if max_chr is not None:
                if max_chr not in queries_on_target:
                    queries_on_target[max_chr] = []
                queries_on_target[max_chr].append(contig)
        return queries_on_target

    def build_query_on_target_association_file(self):
        """
        For each query, get the best matching chromosome and save it to a CSV file.
        Use the order of queries

        :return: content of the file
        """
        query_on_target = self.get_query_on_target_association(with_coords=True)
        content = "Query\tTarget\tStrand\tQ-len\tQ-start\tQ-stop\tT-len\tT-start\tT-stop\n"
        for contig in self.q_order:
            strand = "+"
            if contig in self.q_reversed:
                strand = "-" if self.q_reversed[contig] else "+"
            if contig in query_on_target:
                min_target = str(query_on_target[contig][3])
                if min_target == "-1":
                    min_target = "na"
                max_target = str(query_on_target[contig][4])
                if max_target == "-1":
                    max_target = "na"
                min_query = str(query_on_target[contig][1])
                if min_query == "-1":
                    min_query = "na"
                max_query = str(query_on_target[contig][2])
                if max_query == "-1":
                    max_query = "na"
                chrm = query_on_target[contig][0]
                content += "%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (contig, chrm or "None", strand,
                                                                     str(self.q_contigs[contig]), min_query, max_query,
                                                                     str(self.t_contigs[chrm]), min_target, max_target)
            else:
                content += "%s\t%s\t%s\t%s\tna\tna\tna\tna\tna\n" % (contig, "None", strand,
                                                                     str(self.q_contigs[contig]))
        return content

    def build_list_no_assoc(self, to):
        """
        Build list of queries that match with None target, or the opposite

        :param to: query or target
        :return: content of the file
        """
        index = self.idx_q if to == "query" else self.idx_t
        name, contigs_list, contigs, reversed, abs_start, c_len = Index.load(index)
        contigs_list = set(contigs_list)
        with open(self.paf, "r") as paf:
            for line in paf:
                c_name = line.strip("\n").split("\t")[0 if to == "query" else 5]
                if c_name in contigs_list:
                    contigs_list.remove(c_name)
        return "\n".join(contigs_list) + "\n"

    def _add_percents(self, percents, item):
        """
        Update percents with interval

        :param percents: initial percents
        :type percents: dict
        :param item: interval from IntervalTree
        :type item: Interval
        :return: new percents
        :rtype: dict
        """
        i_count = item.length()
        percents[str(item.data)] += i_count
        percents["-1"] -= i_count
        return percents

    def _remove_overlaps(self, position_idy, percents):
        """
        Remove overlaps between matches on the diagonal

        :param position_idy: matches intervals with associated identity category
        :type position_idy: IntervalTree
        :param percents: Percent of matches for each identity category
        :type percents: dict
        :return: new percents (updated after overlap removing)
        :rtype: dict
        """
        while len(position_idy) > 0:
            item = position_idy.pop()
            start = item.begin
            end = item.end
            cat = item.data
            overlaps = position_idy.overlap(start, end)
            if len(overlaps) > 0:
                has_overlap = False
                for overlap in overlaps:
                    if has_overlap:
                        break
                    o_start = overlap.begin
                    o_end = overlap.end
                    o_cat = overlap.data
                    if not position_idy.containsi(o_start, o_end, o_cat):
                        continue
                    if start < o_start:
                        if end <= o_end:
                            # cccccccccccccc*******
                            # *****ooooooooo[ooooooo]
                            if o_cat < cat:
                                if end < o_end:
                                    # No overlap with the current item, we stay has_overlap as False
                                    position_idy.discard(overlap)
                                    position_idy[end:o_end] = o_cat
                                else:
                                    position_idy.discard(overlap)  # No kept overlap
                            elif o_cat == cat:
                                if end < o_end:
                                    has_overlap = True
                                    position_idy.discard(overlap)
                                    position_idy[start:o_end] = cat
                                else:
                                    position_idy.discard(overlap)  # No kept overlap
                            else:
                                has_overlap = True
                                position_idy.discard(overlap)
                                position_idy[start:o_start] = cat
                                position_idy[o_start:o_end] = o_cat
                        else:  # end > o_end
                            # ccccccccccccccccccc
                            # *****oooooooooo****
                            if o_cat <= cat:
                                position_idy.discard(overlap)  # No kept overlap
                            else:  # o_cat > cat
                                has_overlap = True
                                position_idy.discard(overlap)
                                position_idy[start:o_start] = cat
                                position_idy[o_start:o_end] = o_cat
                                position_idy[o_end:end] = cat
                    elif start == o_start:
                        if end < o_end:
                            # cccccccccccc*******
                            # ooooooooooooooooooo
                            if o_cat < cat:
                                # No overlap with the current item, we stay has_overlap as False
                                position_idy.discard(overlap)
                                position_idy[end:o_end] = o_cat
                            elif o_cat == cat:
                                has_overlap = True
                                position_idy.discard(overlap)
                                position_idy[start:o_end] = cat
                            else:  # o_cat > cat
                                # The overlap just contains current item
                                has_overlap = True
                        elif end == o_end:
                            # ***cccccccccccccccc***
                            # ***oooooooooooooooo***
                            if o_cat <= cat:
                                position_idy.discard(overlap)  # No kept overlap
                            else:
                                # The overlap just contains current item
                                has_overlap = True
                        else:  # end > o_end
                            # ccccccccccccccccccccccccccccc
                            # oooooooooooooooooooo*********
                            if o_cat <= cat:
                                # current item just contains the overlap
                                position_idy.discard(overlap)
                            else:
                                has_overlap = True
                                position_idy.discard(overlap)
                                position_idy[o_start:o_end] = o_cat
                                position_idy[o_end:end] = cat
                    else:  # start > o_start
                        if end <= o_end:
                            # ******ccccccccc*******
                            # ooooooooooooooo[ooooooo]
                            if o_cat < cat:
                                has_overlap=True
                                position_idy.discard(overlap)
                                position_idy[o_start:start] = o_cat
                                position_idy[start:end] = cat
                                if end < o_end:
                                    position_idy[end:o_end] = o_cat
                            else:  # o_cat >= cat
                                # Overlap just contains the item
                                has_overlap = True
                        else: # end > o_end
                            # ******ccccccccccccccccccccc
                            # ooooooooooooooooo**********
                            if o_cat < cat:
                                has_overlap = True
                                position_idy.discard(overlap)
                                position_idy[o_start:start] = o_cat
                                position_idy[start:end] = cat
                            elif o_cat == cat:
                                has_overlap = True
                                position_idy.discard(overlap)
                                position_idy[o_start:end] = cat
                            else:  # o_cat > cat
                                has_overlap = True
                                position_idy[o_end:end] = cat
                if not has_overlap:
                    percents = self._add_percents(percents, item)

            else:
                percents = self._add_percents(percents, item)

        return percents

    def build_summary_stats(self, status_file):
        """
        Get summary of identity

        :return: table with percents by category
        """
        summary_file = self.paf + ".summary"
        self.parse_paf(False, False)
        if self.parsed:
            percents = {"-1": self.len_t}
            position_idy = IntervalTree()

            cats = sorted(self.lines.keys())

            for cat in cats:
                percents[cat] = 0
                for line in self.lines[cat]:
                    start = min(line[0], line[1])
                    end = max(line[0], line[1]) + 1
                    position_idy[start:end] = int(cat)

            percents = self._remove_overlaps(position_idy, percents)

            for cat in percents:
                percents[cat] = percents[cat] / self.len_t * 100

            with open(summary_file, "w") as summary_file:
                summary_file.write(json.dumps(percents))

            os.remove(status_file)
            return percents
        shutil.move(status_file, status_file + ".fail")
        return None

    def get_summary_stats(self):
        """
        Load summary statistics from file

        :return: summary object or None if summary not already built
        :rtype: dict
        """
        summary_file = self.paf + ".summary"
        if os.path.exists(summary_file):
            with open(summary_file, "r") as summary_file:
                txt = summary_file.read()
                return json.loads(txt)
        return None

    def build_query_chr_as_reference(self):
        """
        Assemble query contigs like reference chromosomes

        :return: path of the fasta file
        """
        try:
            if not self.sorted:
                raise Exception("Contigs must be sorted to do that!")
            with open(os.path.join(self.data_dir, ".query")) as query_file:
                query_fasta = query_file.read().strip("\n")
            if not os.path.isfile(query_fasta):
                raise Exception("Query fasta does not exists")
            o_fasta = os.path.join(os.path.dirname(query_fasta), "as_reference_" + os.path.basename(query_fasta))
            if o_fasta.endswith(".gz"):
                o_fasta = o_fasta[:-3]
            if not os.path.exists(o_fasta):
                uncompressed = False
                if query_fasta.endswith(".gz"):
                    uncompressed = True
                    query_fasta = Functions.uncompress(query_fasta)
                query_f = SeqIO.index(query_fasta, "fasta")
                contigs_assoc = self.get_queries_on_target_association()
                mapped_queries = set()
                with open(o_fasta, "w") as out:
                    for target in self.t_order:
                        if target in contigs_assoc:
                            queries = sorted(contigs_assoc[target], key=lambda x: self.q_order.index(x))
                            seq = SeqRecord(Seq(""))
                            for query in queries:
                                mapped_queries.add(query)
                                new_seq = query_f[query]
                                if self.q_reversed[query]:
                                    new_seq = new_seq.reverse_complement()
                                seq += new_seq
                                seq += 100 * "N"
                            seq = seq[:-100]
                            seq.id = seq.name = seq.description = target
                            SeqIO.write(seq, out, "fasta")
                    for contig in self.q_order:
                        if contig not in mapped_queries:
                            seq = query_f[contig]
                            seq.id += "_unaligned"
                            SeqIO.write(seq, out, "fasta")
                query_f.close()
                if uncompressed:
                    os.remove(query_fasta)
            status = "success"
        except Exception:
            o_fasta = "_._"
            status="fail"

        if MODE == "webserver":
            parts = os.path.basename(o_fasta).rsplit(".", 1)
            Functions.send_fasta_ready(mailer=self.mailer,
                                       job_name=self.id_job,
                                       sample_name=parts[0],
                                       ext=parts[1],
                                       compressed=False,
                                       path="download",
                                       status=status)
        return o_fasta



"""
Define formats validators here (for alignment files)

Each validator (main function) has a name which is exactly the name of the format in the aln-formats.yaml file.
Only 1 argument to this function:
- Input file to check

Secondary functions must start with _

Validators for non-mapping files must start with "v_"

Returns True if file is valid, else False
"""

from Bio import AlignIO
import os
import shutil
import traceback
from functools import reduce


def _good_paf_line(parts):
    """
    Check PAF line splitted in many parts

    :param parts: line splitted in many parts
    :type parts: list
    :return: True if line is correctly formatted
    :rtype: bool
    """
    result = len(parts) >= 12 \
        and reduce(lambda x, y: x and y, (z.isdigit() for z in parts[1:4])) \
        and parts[4] in ['+', '-'] \
        and reduce(lambda x, y: x and y, (z.isdigit() for z in parts[6:12])) \
        and 0 <= int(parts[11]) <= 255
    return result


def paf(in_file, n_max=None):
    """
    Paf validator

    :param in_file: paf file to test
    :type in_file: str
    :param n_max: number of lines to test (default: None for all)
    :type n_max: int
    :return: True if valid, else False
    :rtype: bool
    """
    try:
        with open(in_file, "r") as aln:
            n = 0
            for line in aln:
                parts = line.rstrip().split("\t")
                if not _good_paf_line(parts):
                    return False
                n += 1
                if n_max and n >= n_max:
                    break
    except:
        traceback.print_exc()
        return False
    else:
        return True


def _filter_maf(in_file):
    """
    Filter Maf file (remove unused lines)

    :param in_file: maf file to filter
    """
    new_file = in_file + ".new"
    with open(in_file, "r") as inf, open(new_file, "w") as new:
        for line in inf:
            if len(line.rstrip()) == 0 or line[0] in ["#", "a", "s"]:
                new.write(line)
    os.remove(in_file)
    shutil.move(new_file, in_file)


def maf(in_file):
    """
    Maf validator

    :param in_file: maf file to test
    :type in_file: str
    :return: True if valid, else False
    :rtype: bool
    """
    _filter_maf(in_file)
    try:
        maf = AlignIO.parse(in_file, "maf")
        for grp in maf:
            if len(grp) != 2:
                return False
    except:
        return False
    else:
        return True


def v_idx(in_file):
    """
    Index file validator

    :param in_file: index file to test
    :type in_file: str
    :return: True if valid, else False
    :rtype: bool
    """
    try:
        with open(in_file, "r") as inf:
            first_line = inf.readline()
            if "\t" in first_line:
                return False
            must_be_last = False
            for line in inf:
                if must_be_last:
                    return False
                line = line.rstrip()
                if line == "":
                    must_be_last = True
                else:
                    cols = line.split("\t")
                    if len(cols) != 2:
                        return False
                    if not cols[1].isdigit():
                        return False
    except:
        return False
    else:
        return True



#!/usr/bin/env python3

import os
import sys
import shutil
import time
from _datetime import datetime, timedelta
import traceback
import argparse

from dgenies.config_reader import AppConfigReader
from dgenies.lib.functions import Functions

config_reader = AppConfigReader()


def parse_upload_folders(upload_folder, now, max_age, fake=False):
    """
    Parse upload folders and remove too old files and folders

    :param upload_folder: upload folder path
    :type upload_folder: str
    :param now: current timestamp
    :type now: float
    :param max_age: remove all files & folders older than this age. Define it for each category
        (uploads, data, error, ...)
    :type max_age: dict
    :param fake: if True, just print files to delete, without delete them
    :type fake: bool
    """
    for file in os.listdir(upload_folder):
        file = os.path.join(upload_folder, file)
        create_date = os.path.getctime(file)
        age = (now - create_date) / 86400  # Age in days
        if age > max_age["uploads"]:
            try:
                if os.path.isdir(file):
                    print("Removing folder %s..." % file)
                    if not fake:
                        shutil.rmtree(file)
                else:
                    print("Removing file %s..." % file)
                    if not fake:
                        os.remove(file)
            except OSError:
                print(traceback.print_exc(), file=sys.stderr)


def parse_database(app_data, max_age, fake=False):
    """
    Parse database and remove too old jobs (from database and from disk)

    :param app_data: folder where jobs are stored
    :type app_data: str
    :param max_age: remove all files & folders older than this age. Define it for each category
        (uploads, data, error, ...)
    :type max_age: dict
    :param fake: if True, just print files to delete, without delete them
    :type fake: bool
    :return: id jobs which are in the gallery (not removed independently of their age)
    :rtype: list
    """
    from dgenies.database import Job, Gallery
    gallery_jobs = []
    with Job.connect():
        old_jobs = Job.select().where(
                      ((Job.status == "success") & (Job.date_created < datetime.now() - timedelta(days=max_age["data"])))
                      |
                      ((Job.status != "success") & (Job.date_created < datetime.now() - timedelta(days=max_age["error"])))
                  )
        for job in old_jobs:
            id_job = job.id_job
            is_gallery = len(Gallery.select().join(Job).where(Job.id_job == id_job)) > 0
            if is_gallery:
                gallery_jobs.append(id_job)
            else:
                print("Removing job %s..." % id_job)
                data_dir = os.path.join(app_data, id_job)
                if os.path.exists(data_dir) and os.path.isdir(data_dir):
                    if not fake:
                        shutil.rmtree(data_dir)
                else:
                    print("Job %s has no data folder!" % id_job)
                if not fake:
                    job.delete_instance()
    return gallery_jobs


def parse_data_folders(app_data, gallery_jobs, now, max_age, fake=False):
    """
    Parse data folder and remove too old jobs

    :param app_data: folder where jobs are stored
    :param gallery_jobs: id of jobs which are inside the gallery
    :type gallery_jobs: list
    :param now: current timestamp
    :type now: float
    :param max_age: remove all files & folders older than this age. Define it for each category
        (uploads, data, error, ...)
    :type max_age: dict
    :param fake: if True, just print files to delete, without delete them
    :type fake: bool
    :return:
    """
    for file in os.listdir(app_data):
        if file not in gallery_jobs and file not in ["gallery"]:
            file = os.path.join(app_data, file)
            create_date = os.path.getctime(file)
            age = (now - create_date) / 86400  # Age in days
            if age > max_age["data"]:
                try:
                    if os.path.isdir(file):
                        print("Removing folder %s..." % file)
                        if not fake:
                            shutil.rmtree(file)
                    else:
                        print("Removing file %s..." % file)
                        if not fake:
                            os.remove(file)
                except OSError:
                    print(traceback.print_exc())
            elif os.path.isdir(file):
                query_name_file = os.path.join(file, ".query")
                if os.path.exists(query_name_file):
                    with open(query_name_file) as query_file:
                        query_filename = query_file.read().strip("\n")
                        sorted_file = Functions.get_fasta_file(file, "query", True)
                        if not sorted_file.endswith(".sorted"):
                            sorted_file = None
                        if sorted_file is not None:
                            create_date = os.path.getctime(sorted_file)
                            age = (now - create_date) / 86400  # Age in days
                            if age > max_age["fasta_sorted"]:
                                print("Removing fasta file %s..." % sorted_file)
                                if not fake:
                                    os.remove(sorted_file)
                        query_reference = os.path.join(file, "as_reference_" + os.path.basename(query_filename))
                        if os.path.exists(query_reference):
                            create_date = os.path.getctime(query_reference)
                            age = (now - create_date) / 86400  # Age in days
                            if age > max_age["fasta_sorted"]:
                                print("Removing fasta file %s..." % query_reference)
                                if not fake:
                                    os.remove(query_reference)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Clean old jobs and files")
    parser.add_argument('-f', '--fake', type=bool, const=True, nargs="?", required=False, default=False,
                        help="Fake mode: don't really delete the files (ONLY for debug)")
    parser.add_argument("-d", "--max-age", type=int, required=False, help="Max age of jobs to delete", default=7)
    args = parser.parse_args()
    fake = args.fake

    upload_folder = config_reader.upload_folder
    app_data = config_reader.app_data
    now = time.time()

    max_age = {
        "uploads": 1,
        "error": 1,
        "data": args.max_age,
        "fasta_sorted": 1
    }

    print("#########################")
    print("# Parsing Upload folder #")
    print("#########################")
    print("")
    parse_upload_folders(
        upload_folder=upload_folder,
        now=now,
        max_age=max_age,
        fake=fake
    )
    print("")

    print("######################")
    print("# Parsing Jobs in DB #")
    print("######################")
    print("")
    gallery_jobs = parse_database(
        app_data=app_data,
        max_age=max_age,
        fake=fake
    )
    print("")

    print("#######################")
    print("# Parsing Data folder #")
    print("#######################")
    print("")
    parse_data_folders(
        app_data=app_data,
        now=now,
        max_age=max_age,
        fake=fake,
        gallery_jobs=gallery_jobs
    )
    print("")



try:
    import drmaa
except:
    print("Error: unable to import DRMAA module")
from .decorators import Singleton

@Singleton
class DrmaaSession:
    """
    Initialize and close a DRMAA session (for job submission to a cluster)
    """

    def __init__(self):
        self.session = drmaa.Session()
        self.session.initialize()

    def exit(self):
        print("Exiting DRMAA...")
        self.session.exit()



#!/usr/bin/env python3

import os
import re
import shutil
try:
    from dgenies.bin.index import Index
except ImportError:
    from index import Index
from pathlib import Path
from Bio import SeqIO


class Filter:

    """
    Filter of a fasta file: remove too small contigs
    """

    def __init__(self, fasta, index_file, type_f, min_filtered=0, split=False, out_fasta=None, replace_fa=False):
        """

        :param fasta: fasta file path
        :type fasta: str
        :param index_file: index file path
        :type index_file: str
        :param type_f: type of sample (query or target)
        :type type_f: str
        :param min_filtered: minimum number of large contigs to allow filtering
        :type min_filtered: int
        :param split: are contigs split
        :type split: bool
        :param out_fasta: output fasta file path
        :type out_fasta: str
        :param replace_fa: if True, replace fasta file
        :type replace_fa: bool
        """
        self.fasta = fasta
        self.index_file = index_file
        self.type_f = type_f
        self.min_filtered = min_filtered
        self.split = split
        if out_fasta is not None:
            self.out_fasta = out_fasta
        else:
            self.out_fasta = os.path.join(os.path.dirname(self.fasta), "filtered_" + os.path.basename(self.fasta))
        self.replace_fa = replace_fa

    def filter(self):
        """
        Run filter of contigs

        :return: True if success, else False
        :rtype: bool
        """
        f_outs = self._check_filter()
        if len(f_outs) > 0:
            self._filter_out(f_outs=f_outs)
            return True
        return False

    def _check_filter(self):
        """
        Load index of fasta file, and determine contigs which must be removed. Remove them only in the index

        :return: list of contigs which must be removed
        :rtype: list
        """
        # Load contigs:
        name, order, contigs, reversed_c, abs_start, c_len = Index.load(index_file=self.index_file,
                                                                        merge_splits=self.split)

        # Sort contigs:
        contigs_order = sorted(order, key=lambda x: -contigs[x])

        # Find the N90:
        sum_l = 0
        n95_contig = None
        n95_value = 0.95 * c_len
        pos = -1
        len_small_contigs = 0
        len_1_pct = 0.01 * c_len
        for contig in contigs_order:
            pos += 1
            sum_l += contigs[contig]
            if contigs[contig] < len_1_pct:
                len_small_contigs += contigs[contig]
            if sum_l >= n95_value:
                n95_contig = contig

        if self.type_f == "query" and len_small_contigs >= 0.7 * 0.95 * c_len:
            Path(os.path.join(os.path.dirname(self.fasta), ".do-sort")).touch()

        # Min length of contigs
        min_length = 0.05 * contigs[n95_contig]

        f_outs = []

        breakpoint = None

        for contig in contigs_order[pos:]:
            if contigs[contig] < min_length:
                breakpoint = pos
                break
            pos += 1

        if breakpoint is not None:
            f_outs = contigs_order[breakpoint:]
            if len(f_outs) > self.min_filtered:
                with open(os.path.join(os.path.dirname(self.fasta), ".filter-" + self.type_f), "w") as list_f:
                    list_f.write("\n".join(f_outs) + "\n")
                kept = contigs_order[:breakpoint]
                if self.split:
                    f_outs = []
                    name, contigs_order_split, contigs, reversed_c, abs_start_split, c_len_split = \
                        Index.load(index_file=self.index_file, merge_splits=False)
                    kept_s = []
                    for contig in contigs_order_split:
                        match = re.match(r"(.+)_###_\d+", contig)
                        contig_name = contig
                        if match is not None:
                            contig_name = match.group(1)
                        if contig_name in kept:
                            kept_s.append(contig)
                        else:
                            f_outs.append(contig)
                    kept = kept_s
                else:
                    kept.sort(key=lambda k: order.index(k))
                Index.save(index_file=self.index_file,
                           name=name,
                           contigs=contigs,
                           order=kept,
                           reversed_c=reversed_c)
            else:
                f_outs = []

        return f_outs

    def _filter_out(self, f_outs):
        """
        Remove too small contigs from Fasta file

        :param f_outs: contigs which must be filtered out
        :type f_outs: list
        """
        sequences = SeqIO.parse(open(self.fasta), "fasta")
        keeped = (record for record in sequences if record.name not in f_outs)
        with open(self.out_fasta, "w") as out_fa:
            SeqIO.write(keeped, out_fa, "fasta")

        if self.replace_fa:
            os.remove(self.fasta)
            shutil.move(self.out_fasta, self.fasta)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Filter too small contigs")
    parser.add_argument('-f', '--fasta', type=str, required=True, help="Input fasta file")
    parser.add_argument('-i', '--index', type=str, required=True, help="Index file for the fasta file")
    parser.add_argument('-t', '--type', type=str, required=True, choices=["query", "target"],
                        help="Type of fasta: query or target")
    parser.add_argument('-m', '--min-filtered', type=int, required=False, default=0,
                        help="Minimum number of filtered contigs")
    parser.add_argument('-r', "--replace-fasta", action='store_const', const=True, default=False,
                        help="Replace original fasta file")
    parser.add_argument('-s', "--split", action='store_const', const=True, default=False,
                        help="Is fasta split")
    args = parser.parse_args()

    filter_f = Filter(fasta=args.fasta,
                      index_file=args.index,
                      type_f=args.type,
                      min_filtered=args.min_filtered,
                      split=args.split,
                      replace_fa=args.replace_fasta)
    filter_f.filter()



#!/usr/bin/env python3

import os
from flask import Flask
from .config_reader import AppConfigReader
from .lib.crons import Crons

VERSION = "1.3.0"

app = None
app_title = None
APP_DATA = None
config_reader = None
mailer = None
app_folder = None
MODE = "webserver"
DEBUG = False


def launch(mode="webserver", debug=False):
    """
    Launch the application

    :param mode: webserver or standalone
    :type mode: str
    :param debug: True to enable debug mode
    :type debug: bool
    :return: flask app object
    :rtype: Flask
    """
    global app, app_title, app_folder, APP_DATA, config_reader, mailer, MODE, DEBUG
    app_folder = os.path.dirname(os.path.realpath(__file__))

    MODE = mode
    DEBUG = debug

    # Init config reader:
    config_reader = AppConfigReader()

    UPLOAD_FOLDER = config_reader.upload_folder
    APP_DATA = config_reader.app_data

    app_title = "D-GENIES - Dotplot large Genomes in an Interactive, Efficient and Simple way"

    # Init Flask:
    app = Flask(__name__, static_url_path='/static')
    app.config['UPLOAD_FOLDER'] = UPLOAD_FOLDER
    app.config['MAX_CONTENT_LENGTH'] = config_reader.max_upload_file_size
    app.config['SECRET_KEY'] = 'dsqdsq-255sdA-fHfg52-25Asd5'

    # Init mail:
    if MODE == "webserver":
        from .lib.mailer import Mailer
        mailer = Mailer(app)

    # Create data dir if not exists
    if not os.path.exists(config_reader.app_data):
        os.makedirs(config_reader.app_data)

    if config_reader.debug and config_reader.log_dir != "stdout" and not os.path.exists(config_reader.log_dir):
        os.makedirs(config_reader.log_dir)

    # Crons:
    if os.getenv('DISABLE_CRONS') != "True" and MODE == "webserver":
        print("Starting crons...")
        crons = Crons(app_folder, config_reader.debug or os.getenv('LOGS') == "True")
        crons.start_all()

    from dgenies import views

    return app



class Fasta:
    """
    Defines a fasta file: name of the sample, path to the fasta file, type of file (URL or local file), ...
    """

    def __init__(self, name, path, type_f, example=False):
        """

        :param name: sample name
        :type name: str
        :param path: fasta file path
        :type path: str
        :param type_f: type of file (local file or URL)
        :type type_f: str
        :param example: is an example job
        :type example: bool
        """
        self._name = name
        self._path = path
        self._type = type_f
        self._example = example is not False

    def set_path(self, path):
        """
        Set path to the fasta file

        :param path: new path
        :type path: str
        """
        self._path = path

    def get_path(self):
        """
        Get path of the fasta file

        :return: fasta path
        :rtype: str
        """
        return self._path

    def set_name(self, name):
        """
        Set sample name

        :param name: new sample name
        :type name: str
        """
        self._name = name

    def get_name(self):
        """
        Get sample name

        :return: sample name
        :rtype: str
        """
        return self._name

    def get_type(self):
        """
        Get type: URL or local file

        :return: type
        :rtype: str
        """
        return self._type

    def is_example(self):
        """
        Return if current sample is an example data

        :return: current sample is an example data
        :rtype: bool
        """
        return self._example



import os
import sys
from setuptools import setup, find_packages
from pathlib import Path

with open('requirements.txt') as f:
    install_reqs = f.read().strip().split('\n')
version = '1.3.0'

if os.name == "posix":

    config_dir = '/etc/dgenies'
    wsgi_dir = '/var/www/dgenies'
    if '--user' in sys.argv:
        config_dir = os.path.join(str(Path.home()), ".dgenies")
        wsgi_dir = config_dir

    data_files = [(config_dir, ['application.properties']),
                  (config_dir, ['tools.yaml']),
                  (wsgi_dir, ['dgenies.wsgi'])]

    if "readthedoc" in sys.executable:
        data_files = []

    setup(
        name='dgenies',
        version=version,
        packages=find_packages('src'),
        package_dir={'dgenies': 'src/dgenies'},
        include_package_data=True,
        zip_safe=False,
        install_requires=install_reqs,
        data_files=data_files,
        scripts=['src/bin/dgenies'],
    )

else:

    setup(
        name='dgenies',
        version=version,
        packages=find_packages('src'),
        package_dir={'dgenies': 'src/dgenies'},
        include_package_data=True,
        zip_safe=False,
        install_requires=install_reqs,
        data_files=[('.dgenies', ['application.properties'])],
        scripts=['src/bin/dgenies'],
    )



#!/usr/bin/env python3

import sys
import os
import re
from xopen import xopen
import io
from collections import OrderedDict


class Splitter:

    """
    Split large contigs in smaller ones
    """

    def __init__(self, input_f, name_f, output_f, size_c=10000000, query_index="query_split.idx", debug=False):
        """

        :param input_f: input fasta file path
        :type input_f: str
        :param name_f: sample name
        :type name_f: str
        :param output_f: output fasta file path
        :type output_f: str
        :param size_c: size of split contigs
        :type size_c: int
        :param query_index: index file path for query
        :type query_index: str
        :param debug: True to enable debug mode
        :type debug: bool
        """
        self.input_f = input_f
        self.name_f = name_f
        self.size_c = size_c
        self.output_f = output_f
        self.input_gz = input_f.endswith(".gz")
        self.output_gz = output_f.endswith(".gz")
        self.out_dir = os.path.dirname(output_f)
        self.index_file = os.path.join(self.out_dir, query_index)
        self.nb_contigs = 0
        self.debug = debug

    def split(self):
        """
        Split contigs in smaller ones staff

        :return: True if the input Fasta is correct, else False
        """
        has_header = False
        next_header = False  # True if next line must be a header line
        with (xopen(self.input_f, mode="r") if self.input_gz else open(self.input_f, mode="r")) as fasta, \
            (xopen(self.output_f, mode="w") if self.output_gz else open(self.output_f, mode="w")) as enc, \
                open(self.index_file, mode="w") as index_f:
            index_f.write(self.name_f + "\n")
            chr_name = None
            fasta_str = ""
            nb_line = 0
            for line in fasta:
                nb_line += 1
                line = line.strip("\n")
                if re.match(r"^>.+", line) is not None:
                    has_header = True
                    next_header = False
                    if chr_name is not None and len(fasta_str) > 0:
                        self.nb_contigs += 1
                        self.flush_contig(fasta_str, chr_name, self.size_c, enc, index_f)
                    elif chr_name is not None:
                        return False, "Error: contig is empty: %s" % chr_name
                    chr_name = re.split("\s", line[1:])[0]
                    fasta_str = ""
                    if self.debug:
                        print("Parsing contig \"%s\"... " % chr_name, end="")
                elif len(line) > 0:
                    if next_header or re.match(r"^[ATGCKMRYSWBVHDXN.\-]+$", line.upper()) is None:
                        if next_header:
                            return False, "Error: new header line expected at line %d" % nb_line
                        return False, "Error: invalid sequence at line %d" % nb_line
                    fasta_str += line
                elif len(line) == 0:
                    next_header = True
            self.nb_contigs += 1
            self.flush_contig(fasta_str, chr_name, self.size_c, enc, index_f)
        return has_header, ""

    @staticmethod
    def write_contig(name, fasta, o_file):
        o_file.write(">%s\n" % name)
        f_len = len(fasta)
        i = 0
        while i<f_len:
            j = min (i+60, f_len)
            o_file.write(fasta[i:j] + "\n")
            i = j
        o_file.write("\n")

    @staticmethod
    def split_contig(name, sequence, block_sizes):
        seq_len = len(sequence)
        contigs = OrderedDict()
        if seq_len < block_sizes + block_sizes * 0.1:  # Only one block size
            contigs[name] = sequence
            return contigs
        i = 0
        n = 1
        while i < seq_len:
            j = i + block_sizes
            if (j > seq_len) or ((i + (block_sizes * 0.1)) >= seq_len):
                j = seq_len
            contigs[name + ("_###_%d" % n)] = sequence[i:j]
            i = j
            n += 1
        return contigs

    def flush_contig(self, fasta_str, chr_name, size_c, enc, index_f):
        if len(fasta_str) > 0 and chr_name is not None:
            contigs = self.split_contig(chr_name, fasta_str, size_c)
            for name, seq in contigs.items():
                self.write_contig(name, seq, enc)
                index_f.write("%s\t%d\n" % (name, len(seq)))
            nb_contigs = len(contigs)
            if self.debug:
                if nb_contigs == 1:
                    print("Keeped!")
                else:
                    print("Splited in %d contigs!" % nb_contigs)


def parse_args():
    import argparse

    parser = argparse.ArgumentParser(description="Split huge contigs")
    parser.add_argument('-i', '--input', type=str, required=True, help="Input fasta file")
    parser.add_argument('-n', '--name', type=str, required=True, help="Input fasta name")
    parser.add_argument('-s', '--size', type=int, required=False, default=10, help="Max size of contigs (Mb)")
    parser.add_argument('-o', '--output', type=str, required=True, help="Output fasta file")
    args = parser.parse_args()

    if args.size < 0:
        parser.error("Max size of contigs must be positive")

    return args


def __main__():
    args = parse_args()
    splitter = Splitter(args.input, args.name, args.output, args.size * 1000000)
    return not splitter.split()


if __name__ == '__main__':
    sys.exit(__main__())



class Singleton:
    """
    Define a singleton (design pattern)
    """

    def __init__(self, klass):
        self.klass = klass
        self.instance = None

    def __call__(self, *args, **kwds):
        if self.instance is None:
            self.instance = self.klass(*args, **kwds)
        return self.instance



import requests
import json
import os
import threading
from dgenies.config_reader import AppConfigReader


class Latest:

    """
    Search latest version
    """

    def __init__(self):
        self.latest = ""
        self.win32 = ""
        config = AppConfigReader()
        self._save_latest = os.path.join(config.config_dir, ".latest")
        self.load()

    def load(self):
        """
        Load latest version: use cached version (if any) and then sync with Github
        """
        if os.path.exists(self._save_latest):
            with open(self._save_latest, "r") as latest_f:
                self.latest = latest_f.readline().rstrip()
                self.win32 = latest_f.readline().rstrip()
            if self.latest == "" or self.win32 == "":
                self.update()
            else:
                self.update_async()
        else:
            self.update()

    def update_async(self):
        """
        Update latest version asynchronously
        """
        thread = threading.Timer(1, self.update)
        thread.start()

    def update(self):
        """
        Get latest version from Github
        """
        try:
            call = requests.get("https://api.github.com/repos/genotoul-bioinfo/dgenies/releases/latest")
            if call.ok:
                release = json.loads(call.content.decode("utf-8"))
                if "tag_name" in release:
                    self.latest = release["tag_name"][1:]
                    for asset in release["assets"]:
                        if asset["name"].endswith(".exe"):
                            self.win32 = asset["browser_download_url"]
                            break
        except ConnectionError:
            pass
        else:
            self._write_update()

    def _write_update(self):
        """
        Save latest version to a file
        """
        if self.latest != "" or self.win32 != "":
            with open(self._save_latest, "w") as latest_f:
                latest_f.write("\n".join([self.latest, self.win32]))



import logging
#logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')

from logging.config import dictConfig

dictConfig({
    'version': 1,
    'formatters': {
        'default': {
            'format': '[%(asctime)s] %(levelname)s:%(module)s: %(message)s',
        },
        'flask': {
            'format': '%(message)s',
        }
    },
    'handlers': {
        'wsgi': {
            'class': 'logging.StreamHandler',
            'stream': 'ext://flask.logging.wsgi_errors_stream',
            'formatter': 'flask'
        },
        'stderr-handler': {
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stderr',
            'formatter': 'default'
        },
        'dgenies-handler': {
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'filename': 'logs/dgenies.log',
            'when': 'W6',
            'interval': 1,
            'formatter': 'default'
        },
        'local-scheduler-handler': {
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'filename': 'logs/local-scheduler.log',
            'when': 'W6',
            'interval': 1,
            'formatter': 'default'
        }
    },
#    'root': {
#        'level': 'INFO',
#        'handlers': ['dgenies-app']
#    },
    'loggers': {
        'flask.app': {
            'level': 'INFO',
            'handlers': ['wsgi']
        },
        'dgenies': {
            'level': 'INFO',
            'handlers': ['stderr-handler', 'dgenies-handler']
        },
        'local_scheduler': {
            'level': 'INFO',
            'handlers': ['stderr-handler', 'local-scheduler-handler']
        }
    },
})


import os
import random
import string
import shutil
import sys
import re
import traceback
from inspect import getmembers, isfunction
from collections import OrderedDict
from Bio import SeqIO
from jinja2 import Template
from xopen import xopen
from dgenies.config_reader import AppConfigReader
import dgenies.lib.validators as validators

ALLOWED_EXTENSIONS = {"fasta": ['fa', 'fasta', 'fna', 'fa.gz', 'fasta.gz', 'fna.gz'],
                      "idx": ['idx'],
                      "map": [o[0] for o in getmembers(validators) if isfunction(o[1]) and not o[0].startswith("_") and
                              not o[0].startswith("v_")],
                      "backup": ['tar', 'tar.gz']}
# map: all functions of validators which does not starts with an underscore.


class Functions:

    """
    General functions
    """

    config = AppConfigReader()

    @staticmethod
    def allowed_file(filename, file_formats=("fasta",)):
        """
        Check whether a file has a valid format

        :param filename: file path
        :param file_formats: accepted file formats
        :return: True if valid format, else False
        """
        for file_format in file_formats:
            if '.' in filename and \
                   (filename.rsplit('.', 1)[1].lower() in ALLOWED_EXTENSIONS[file_format]
                    or ".".join(filename.rsplit('.', 2)[1:]).lower() in ALLOWED_EXTENSIONS[file_format]):
                return True
        return False

    @staticmethod
    def random_string(s_len):
        """
        Generate a random string

        :param s_len: length of the string to generate
        :type s_len: int
        :return: the random string
        :rtype: str
        """
        return ''.join([random.choice(string.ascii_letters + string.digits) for n in range(s_len)])

    @staticmethod
    def get_valid_uploaded_filename(filename, folder):
        """
        Check whether uploaded file already exists. If yes, rename it

        :param filename: uploaded file
        :type filename: str
        :param folder: folder into save the file
        :type folder: str
        :return: unique filename
        :rtype: str
        """
        file_query_s = os.path.join(folder, filename)
        i = 2
        filename_orig = filename
        while os.path.exists(file_query_s):
            filename = str(i) + "_" + filename_orig
            file_query_s = os.path.join(folder, filename)
            i += 1
        return filename

    @staticmethod
    def __get_do_sort(fasta, is_sorted):
        """
        Check whether query must be sorted (False if already done)

        :param fasta: fasta file
        :type fasta: str
        :param is_sorted: True if it's sorted
        :type is_sorted: bool
        :return: do sort
        :rtype: bool
        """
        do_sort = False
        if is_sorted:
            do_sort = True
            if fasta.endswith(".sorted"):
                do_sort = False
        return do_sort

    @staticmethod
    def get_fasta_file(res_dir, type_f, is_sorted):
        """
        Get fasta file path

        :param res_dir: job results directory
        :type res_dir: str
        :param type_f: type of file (query or target)
        :type type_f: str
        :param is_sorted: is fasta sorted
        :type is_sorted: bool
        :return: fasta file path
        :rtype: str
        """
        fasta_file = None
        try:
            with open(os.path.join(res_dir, "." + type_f), "r") as save_name:
                fasta_file = save_name.readline().strip("\n")
        except IOError:
            print(res_dir + ": Unable to load saved name for " + type_f, file=sys.stderr)
            pass
        if fasta_file is not None and os.path.exists(fasta_file):
            fasta_file_uc = fasta_file
            if fasta_file.endswith(".gz"):
                fasta_file_uc = fasta_file[:-3]
            if is_sorted:
                sorted_fasta = fasta_file_uc + ".sorted"
                if os.path.exists(sorted_fasta):
                    fasta_file = sorted_fasta
                else:
                    sorted_fasta = fasta_file_uc + ".gz.sorted"
                    if os.path.exists(sorted_fasta):
                        fasta_file = sorted_fasta

        return fasta_file

    @staticmethod
    def uncompress(filename):
        """
        Uncompress a gzipped file

        :param filename: gzipped file
        :type filename: str
        :return: path of the uncompressed file
        :rtype: str
        """
        try:
            uncompressed = filename.rsplit('.', 1)[0]
            parts = uncompressed.rsplit("/", 1)
            file_path = parts[0]
            basename = parts[1]
            n = 2
            while os.path.exists(uncompressed):
                uncompressed = "%s/%d_%s" % (file_path, n, basename)
                n += 1
            with xopen(filename, "rb") as infile, open(uncompressed, "wb") as outfile:
                outfile.write(infile.read())
            return uncompressed
        except Exception as e:
            print(traceback.format_exc())
            return None

    @staticmethod
    def compress(filename):
        """
        Compress a file with gzip

        :param filename: file to compress
        :type filename: str
        :return: path of the compressed file
        :rtype: str
        """
        try:
            if not filename.endswith(".gz") and not filename.endswith(".gz.sorted"):
                compressed = filename + ".gz" if not filename.endswith(".sorted") else filename[:-7] + ".gz.sorted"
                parts = compressed.rsplit("/", 1)
                file_path = parts[0]
                basename = parts[1]
                n = 2
                while os.path.exists(compressed):
                    compressed = "%s/%d_%s" % (file_path, n, basename)
                    n += 1
                with open(filename, "rb") as infile, xopen(compressed, "wb") as outfile:
                    shutil.copyfileobj(infile, outfile)
                os.remove(filename)
                return compressed
            return filename
        except Exception as e:
            print(traceback.format_exc())
            return None

    @staticmethod
    def read_index(index_file):
        """
        Load index of query or target

        :param index_file: index file path
        :type index_file: str
        :return:
            * [0] index (size of each chromosome) {dict}
            * [1] sample name {str}
        :rtype: (dict, str)
        """
        index = OrderedDict()
        with open(index_file, "r") as index_f:
            # Sample name without special chars:
            sample_name = re.sub('[^A-Za-z0-9_\-.]+', '', index_f.readline().strip("\n").replace(" ", "_"))
            for line in index_f:
                if line != "":
                    parts = line.strip("\n").split("\t")
                    name = parts[0]
                    lenght = int(parts[1])
                    to_reverse = parts[2] == "1" if len(parts) >= 3 else False
                    index[name] = {
                        "length": lenght,
                        "to_reverse": to_reverse
                    }
        return index, sample_name

    @staticmethod
    def get_mail_for_job(id_job):
        """
        Retrieve associated mail for a job

        :param id_job: job id
        :type id_job: int
        :return: associated mail address
        :rtype: str
        """
        from dgenies.database import Job
        with Job.connect():
            j1 = Job.get(Job.id_job == id_job)
            return j1.email

    @staticmethod
    def send_fasta_ready(mailer, job_name, sample_name, compressed=False, path="fasta-query", status="success",
                         ext="fasta"):
        """
        Send link to fasta file when treatment ended

        :param mailer: mailer object
        :type mailer: Mailer
        :param job_name: job id
        :type job_name: str
        :param sample_name: sample name
        :type sample_name: str
        :param compressed: is a compressed fasta file
        :type compressed: bool
        :param path: fasta path
        :type path: str
        :param status: treatment status
        :type status: str
        :param ext: file extension
        :type ext: str
        """
        web_url = Functions.config.web_url
        with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "mail_templates", "dl_fasta.html")) \
                as t_file:
            template = Template(t_file.read())
            message_html = template.render(job_name=job_name, status=status, url_base=web_url,
                                           sample_name=sample_name, compressed=compressed, path=path, ext=ext)
        message = "D-Genies\n\n" \
                  "Job %s - Download fasta\n\n" % job_name
        message += "Query fasta file for job %s (query: %s) is ready to download.\n" % (job_name, sample_name)
        message += "You can click on the link below to download it:\n\n"
        message += "%s/fasta-query/%s/%s" % (web_url, job_name, sample_name + ".fasta" + (".gz" if compressed else ""))
        mailer.send_mail([Functions.get_mail_for_job(job_name)], "Job %s - Download fasta" % job_name, message,
                         message_html)



    @staticmethod
    def sort_fasta(job_name, fasta_file, index_file, lock_file, compress=False, mailer=None, mode="webserver"):
        """
        Sort fasta file according to the sorted index file

        :param job_name: job id
        :type job_name: str
        :param fasta_file: fasta file path
        :type fasta_file: str
        :param index_file: index file path
        :type index_file: str
        :param lock_file: lock file path
        :type lock_file: str
        :param compress: compress result fasta file
        :type compress: bool
        :param mailer: mailer object (to send mail)
        :type mailer: Mailer
        :param mode: webserver or standalone
        :type mode: str
        """
        index, sample_name = Functions.read_index(index_file)
        is_compressed = fasta_file.endswith(".gz")
        if is_compressed:
            fasta_file = Functions.uncompress(fasta_file)
        fasta_file_o = fasta_file + ".sorted"
        seq = SeqIO.index(fasta_file, "fasta")
        with open(fasta_file_o, "w") as fasta_out:
            for name, props in index.items():
                sequence = seq[name]
                if props["to_reverse"]:
                    s_id = sequence.id
                    s_name = sequence.name
                    s_description = sequence.description
                    sequence = sequence.reverse_complement()
                    sequence.id = s_id
                    sequence.name = s_name
                    sequence.description = s_description
                SeqIO.write(sequence, fasta_out, "fasta")
        seq.close()
        if is_compressed:
            os.remove(fasta_file)
        if compress:
            Functions.compress(fasta_file_o)
        os.remove(lock_file)
        if mode == "webserver" and mailer is not None and not os.path.exists(lock_file + ".pending"):
            Functions.send_fasta_ready(mailer, job_name, sample_name, compress)

    @staticmethod
    def compress_and_send_mail(job_name, fasta_file, index_file, lock_file, mailer):
        """
        Compress fasta file and the send mail with its link to the client

        :param job_name: job id
        :type job_name: str
        :param fasta_file: fasta file path
        :type fasta_file: str
        :param index_file: index file path
        :type index_file: str
        :param lock_file: lock file path
        :type lock_file: str
        :param mailer: mailer object (to send mail)
        :type mailer: Mailer
        """
        Functions.compress(fasta_file)
        os.remove(lock_file)
        index, sample_name = Functions.read_index(index_file)
        Functions.send_fasta_ready(mailer, job_name, sample_name, True)

    @staticmethod
    def get_readable_size(size, nb_after_coma=1, base="B"):
        """
        Get human readable size from a given size in bytes

        :param size: size in bytes
        :type size: int
        :param nb_after_coma: number of digits after coma
        :type nb_after_coma: int
        :param base: base unit of size, must be either "B", "KiB", "MiB" or "GiB"
        :type nb_after_coma: str
        :return: size, human readable
        :rtype: str
        """
        print(size)
        units = ["B", "KiB", "MiB", "GiB"]
        i = units.index(base)
        while size >= 1024 and i < (len(units)-1):
            size /= 1024.0
            i += 1
        return str("%." + str(nb_after_coma) + "f %s") % (size, units[i])

    @staticmethod
    def get_readable_time(seconds):
        """
        Get human readable time

        :param seconds: time in seconds
        :type seconds: int
        :return: time, human readable
        :rtype: str
        """
        time_r = "%d s" % seconds
        if seconds >= 60:
            minutes = seconds // 60
            seconds = seconds - (minutes * 60)
            time_r = "%d min %d s" % (minutes, seconds)
            if minutes >= 60:
                hours = minutes // 60
                minutes = minutes - (hours * 60)
                time_r = "%d h %d min %d s" % (hours, minutes, seconds)
        return time_r

    @staticmethod
    def get_gallery_items():
        """
        Get list of items from the gallery

        :return: list of item of the gallery. Each item is a dict with 7 keys:

            * `name` : name of the job
            * `id_job` : id of the job
            * `picture` : illustrating picture filename (located in gallery folder of the data folder)
            * `query` : query specie name
            * `target` : target specie name
            * `mem_peak` : max memory used for the run (human readable)
            * `time_elapsed` : time elapsed for the run (human readable)
        :rtype: list of dict
        """
        from dgenies.database import Gallery
        items = []
        for item in Gallery.select():
            items.append({
                "name": item.name,
                "id_job": item.job.id_job,
                "picture": item.picture,
                "query": item.query,
                "target": item.target,
                "mem_peak": Functions.get_readable_size(item.job.mem_peak, base="KiB"),
                "time_elapsed": Functions.get_readable_time(item.job.time_elapsed)
            })
        return items

    @staticmethod
    def is_in_gallery(id_job, mode="webserver"):
        """
        Check whether a job is in the gallery

        :param id_job: job id
        :type id_job: str
        :param mode: webserver or standalone
        :type mode: str
        :return: True if job is in the gallery, else False
        :rtype: bool
        """
        if mode == "webserver":
            from dgenies.database import Gallery, Job
            from peewee import DoesNotExist
            try:
                return len(Gallery.select().where(Gallery.job == Job.get(id_job=id_job))) > 0
            except DoesNotExist:
                return False
        return False

    @staticmethod
    def _get_jobs_list():
        """
        Get list of jobs

        :return: list of valid jobs
        :rtype: list
        """
        all_jobs = os.listdir(Functions.config.app_data)
        valid_jobs = []
        for job in all_jobs:
            job_path = os.path.join(Functions.config.app_data, job)
            if os.path.isfile(os.path.join(job_path, "map.paf")) and \
                    os.path.isfile(os.path.join(job_path, "target.idx")) and \
                    os.path.isfile(os.path.join(job_path, "query.idx")) and \
                    os.path.isfile(os.path.join(job_path, ".valid")):
                valid_jobs.append(job)
        return valid_jobs

    @staticmethod
    def get_list_all_jobs(mode="webserver"):
        """
        Get list of all jobs

        :param mode: webserver or standalone
        :type mode: str
        :return: list of all jobs in standalone mode. Empty list in webserver mode
        :rtype: list
        """
        if mode == "webserver":
            return []  # Don't give the list in webserver as it's multi-user
        all_jobs = Functions._get_jobs_list()
        if "gallery" in all_jobs:
            all_jobs.remove("gallery")
        return sorted(all_jobs, key=lambda x: x.lower())

    @staticmethod
    def query_fasta_file_exists(res_dir):
        """
        Check if a fasta file exists

        :param res_dir: job result directory
        :type res_dir: str
        :return: True if file exists and is a regular file, else False
        :rtype: bool
        """
        fasta_file = os.path.join(res_dir, ".query")
        return os.path.exists(fasta_file) and os.path.isfile(fasta_file)



import os
import sys
import re
import inspect
from pathlib import Path
import yaml
from dgenies.lib.decorators import Singleton
from dgenies.lib import parsers


class Tool:

    def __init__(self, name, exec, command_line, all_vs_all, max_memory, label=None, threads=1, exec_cluster=None,
                 threads_cluster=None, parser=None, split_before=False, help=None, order=None, options=None):
        """
        Create a new tool

        :param command_line: command line to launch the tool
        :param all_vs_all: command line in all_vs_all mode (None if not available for the tool)
        :param max_memory: max memory the tool is supposed to use (ex: 40G) - for cluster submissions
        :param label: Name to display for user
        :param parser: name of the function in dgenies.lib.functions to launch after mapping to have a correct PAF out
            file
        :param split_before: True to split contigs before mapping
        :type split_before: bool
        :param help: help message to show in run form
        :param order: order to show in run mode
        :param options: list of options for the tool
        """
        # Name
        self.name = name

        # Label
        if label:
            self.label = label
        else:
            self.label = name

        # Exec
        if exec == "default":
            self.exec = os.path.join(os.path.dirname(inspect.getfile(self.__class__)), "bin", self.name)
        else:
            self.exec = exec.replace("###SYSEXEC###", os.path.dirname(sys.executable))
        if exec_cluster is None or exec_cluster == "default":
            self.exec_cluster = exec
        else:
            self.exec_cluster = exec_cluster.replace("###SYSEXEC###", os.path.dirname(sys.executable))

        # Command line:
        if "{exe}" in command_line and "{target}" in command_line and "{query}" in command_line and "{out}" \
                in command_line:
            self.command_line = command_line
        else:
            raise ValueError("Tools: command_line must contains at least {exe}, {target}, {query} and {out} tags")

        # All_vs_all:
        if all_vs_all is None or ("{exe}" in all_vs_all and "{target}" in all_vs_all and "{out}" in all_vs_all):
            self.all_vs_all = all_vs_all
        else:
            raise ValueError("Tools: all_vs_all must contains at least {exe}, {target} and {out} tags")

        # Max memory:
        if max_memory is None or isinstance(max_memory, int):
            self.max_memory = max_memory
        else:
            raise ValueError("Tools: max_memory must be an integer, or !!null")

        # Threads:
        if isinstance(threads, int):
            self.threads = threads
        else:
            raise ValueError("Tools: threads must be an integer")
        if threads_cluster is None or isinstance(threads_cluster, int):
            if threads_cluster is None:
                self.threads_cluster = self.threads
            else:
                self.threads_cluster = threads_cluster
        else:
            raise ValueError("Tools: threads_cluster must be an integer, or !!null")

        # Parser:
        if parser is None or hasattr(parsers, parser):
            self.parser = parser
        else:
            raise ValueError("Tools: parser %s is not defines in dgenies.lib.parsers!" % parser)

        # split_before:
        if isinstance(split_before, bool):
            self.split_before = split_before
        else:
            raise ValueError("Tools: split_before must be a boolean (True or False)")

        # Help:
        self.help = help

        # Order:
        if order is None or isinstance(order, int):
            if order is None:
                self.order = 1000
            else:
                self.order = order

        # Options
        if options is None or isinstance(options, list):
            self.options = options
        else:
            raise ValueError("Tools: options must be a yaml list")



@Singleton
class Tools:
    """
    Load (from yaml file) and store available alignment tools
    """

    def __init__(self):
        self.tools = {}
        self.load_yaml(trusted=True)

    def load_yaml(self, trusted=False):
        app_dir = os.path.dirname(inspect.getfile(self.__class__))
        yaml_file = None
        config_file_search = [os.path.join(os.path.abspath(os.sep), "dgenies", "tools.yaml"),
                              "/etc/dgenies/tools.yaml",
                              "/etc/dgenies/tools.yaml.local",
                              os.path.join(str(Path.home()), ".dgenies", "tools.yaml"),
                              os.path.join(str(Path.home()), ".dgenies", "tools.yaml.local"),
                              os.path.join(app_dir, '..', 'etc', 'dgenies', 'tools.yaml')]

        if os.name == "nt":
            config_file_search.insert(1, os.path.join(sys.executable, '..', "tools.yaml"))
            config_file_search.insert(1, os.path.join(sys.executable, '..', "tools.yaml.local"))

        config_file_search.append(os.path.join(app_dir, "tools-dev.yaml"))
        config_file_search.append(os.path.join(app_dir, "tools-dev.yaml.local"))

        for my_config_file in reversed(config_file_search):
            if os.path.exists(my_config_file):
                yaml_file = my_config_file
                break
        if yaml_file is None:
            raise FileNotFoundError("ERROR: tools.yaml not found.")

        with open(yaml_file, "r") as yml_f:
            tools_dict = yaml.load(yml_f, Loader=yaml.FullLoader if trusted else yaml.SafeLoader)
            tools = {}
            for name, props in tools_dict.items():
                tools[name] = Tool(name=name, **props)
            self.tools.update(tools)



from dgenies import MODE

import os
from dgenies.config_reader import AppConfigReader
from datetime import datetime

config = AppConfigReader()

if MODE == "webserver":
    from peewee import SqliteDatabase, Model, CharField, IntegerField, DateTimeField, BooleanField, MySQLDatabase, \
    OperationalError, ForeignKeyField, __exception_wrapper__

    db_url = config.database_url
    db_type = config.database_type

    # restore RetryOperationalError from peewee 2.10.x
    # https://github.com/coleifer/peewee/issues/1472
    class RetryOperationalError(object):

        def execute_sql(self, sql, params=None, commit=True):
            try:
                cursor = super(RetryOperationalError, self).execute_sql(
                    sql, params, commit)
            except OperationalError:
                if not self.is_closed():
                    self.close()
                with __exception_wrapper__:
                    cursor = self.cursor()
                    cursor.execute(sql, params or ())
                    if commit and not self.in_transaction():
                        self.commit()
            return cursor


    class MyRetryDB(RetryOperationalError, MySQLDatabase):
        pass


    if db_type == "sqlite":
        db = SqliteDatabase(db_url)
    elif db_type == "mysql":
        db = MyRetryDB(host=config.database_url, port=config.database_port, user=config.database_user,
                       passwd=config.database_password, database=config.database_db)
    else:
        raise Exception("Unsupported database type: " + db_type)


    class Database:

        nb_open = 0

        def __init__(self):
            pass

        def __enter__(self):
            Database.nb_open += 1
            try:
                db.connect()
            except OperationalError:
                pass

        def __exit__(self, exc_type, exc_val, exc_tb):
            Database.nb_open -= 1
            if Database.nb_open == 0:
                db.close()


    class BaseModel(Model):

        class Meta:
            database = db

        @classmethod
        def connect(cls):
            return Database()


    class Job(BaseModel):
        id_job = CharField(max_length=50, unique=True)
        email = CharField()
        id_process = IntegerField(null=True)
        batch_type = CharField(max_length=20, default="local")
        status = CharField(max_length=20, default="submitted")
        date_created = DateTimeField()
        error = CharField(default="")
        mem_peak = IntegerField(null=True)
        time_elapsed = IntegerField(null=True)
        tool = CharField(default="minimap2", max_length=50, null=True)
        options = CharField(max_length=127, null=True)


    class Gallery(BaseModel):
        job = ForeignKeyField(Job)
        name = CharField()
        query = CharField()
        target = CharField()
        picture = CharField()


    class Session(BaseModel):
        s_id = CharField(max_length=20, unique=True)
        date_created = DateTimeField()
        upload_folder = CharField(max_length=20)
        last_ping = DateTimeField()
        status = CharField(default="reset")
        keep_active = BooleanField(default=False)  # Uploads made by the server must be keep active

        @classmethod
        def new(cls, keep_active=False):
            from dgenies.lib.functions import Functions
            my_s_id = Functions.random_string(20)
            while len(cls.select().where(cls.s_id == my_s_id)) > 0:
                my_s_id = Functions.random_string(20)
            upload_folder = Functions.random_string(20)
            tmp_dir = config.upload_folder
            upload_folder_path = os.path.join(tmp_dir, upload_folder)
            while os.path.exists(upload_folder_path):
                upload_folder = Functions.random_string(20)
                upload_folder_path = os.path.join(tmp_dir, upload_folder)
            cls.create(s_id=my_s_id, date_created=datetime.now(), upload_folder=upload_folder, last_ping=datetime.now(),
                       keep_active=keep_active)
            return my_s_id

        def ask_for_upload(self, change_status=False):
            all_asked = Session.select().where((Session.status == "pending") | (Session.status == "active")).\
                order_by(Session.date_created)
            nb_asked = len(all_asked)
            if self.status != "reset":
                change_status = False

            status = "pending"
            if self.status == "active" or (change_status and nb_asked < 5):
                status = "active"

            self.status = status
            self.last_ping = datetime.now()
            self.save()

            return self.status == "active"

        def ping(self):
            self.last_ping = datetime.now()
            self.save()


    if config.analytics_enabled:

        class Analytics(BaseModel):
            id_job = CharField(max_length=50, default="unknown")
            date_created = DateTimeField()
            target_size = IntegerField()
            query_size = IntegerField(null=True)
            mail_client = CharField()
            batch_type = CharField(max_length=20)
            job_type = CharField(max_length=5, default="unk")
            status = CharField(max_length=20, default="unknown")
            tool = CharField(default="undefined", max_length=50, null=True)

        if not Analytics.table_exists():
            Analytics.create_table()

    if not Job.table_exists():
        Job.create_table()

    if not Gallery.table_exists():
        Gallery.create_table()

    if not Session.table_exists():
        Session.create_table()

else:

    class Database:

        nb_open = 0

        def __init__(self):
            pass

        def __enter__(self):
            pass

        def __exit__(self, exc_type, exc_val, exc_tb):
            pass

    class Job:

        @classmethod
        def connect(cls):
            return Database()